# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The simplest use of the tools in the dataimport module is to create a wizard which 
uses a session-maker object and a class and imports data from a file or 
by raw tabular entry.
"""

from wizard import ImportWizard, \
    ImportDataSource, \
    ImportDataPreview, \
    ImportIntroPage
